@@ -27,16 +27,23 @@
 #include "interface.h"
 #include "rules.h"
 
-#define FONT_SCALE 0.9
+#define NORMAL_FONT_SCALE 0.9
+#define PAINTED_FONT_SCALE 0.6
+#define PAINTED_ALPHA 0.7
+#define NORMAL_ALPHA 1.0
 #define TAG_OFFSET 0.75
 #define TAG_RADIUS 0.25
 #define HINT_FLASHES 6
+#define HINT_DISABLED 0
 #define HINT_INTERVAL 500
 
+static void hitori_cancel_hinting (Hitori *hitori);
+
 /* Declarations for GtkBuilder */
-gboolean hitori_expose_cb (GtkWidget *drawing_area, GdkEventExpose *event, Hitori *hitori);
+gboolean hitori_draw_cb (GtkWidget *drawing_area, cairo_t *cr, Hitori *hitori);
 gboolean hitori_button_release_cb (GtkWidget *drawing_area, GdkEventButton *event, Hitori *hitori);
 void hitori_destroy_cb (GtkWindow *window, Hitori *hitori);
+void hitori_window_state_event_cb (GtkWindow *window, GdkEventWindowState *event, Hitori *hitori);
 void hitori_new_game_cb (GtkAction *action, Hitori *hitori);
 void hitori_hint_cb (GtkAction *action, Hitori *hitori);
 void hitori_undo_cb (GtkAction *action, Hitori *hitori);
@@ -51,6 +58,7 @@ hitori_create_interface (Hitori *hitori)
 {
 	GError *error = NULL;
 	GtkBuilder *builder;
+	GtkStyleContext *style_context;
 
 	builder = gtk_builder_new ();
 
@@ -82,37 +90,35 @@ hitori_create_interface (Hitori *hitori)
 	hitori->undo_action = GTK_ACTION (gtk_builder_get_object (builder, "undo_menu"));
 	hitori->redo_action = GTK_ACTION (gtk_builder_get_object (builder, "redo_menu"));
 	hitori->hint_action = GTK_ACTION (gtk_builder_get_object (builder, "hint_menu"));
+	hitori->timer_label = GTK_LABEL (gtk_builder_get_object (builder, "hitori_timer"));
 
 	g_object_unref (builder);
 
+	/* Set up font descriptions for the drawing area */
+	style_context = gtk_widget_get_style_context (hitori->drawing_area);
+	hitori->normal_font_desc = pango_font_description_copy (gtk_style_context_get_font (style_context, 0));
+	hitori->painted_font_desc = pango_font_description_copy (gtk_style_context_get_font (style_context, 0));
+
+	/* Reset the timer */
+	hitori_reset_timer (hitori);
+
 	return hitori->window;
 }
 
-void
-hitori_draw_board (Hitori *hitori, cairo_t *cr, gboolean check_win)
+gboolean
+hitori_draw_cb (GtkWidget *drawing_area, cairo_t *cr, Hitori *hitori)
 {
 	gint area_width, area_height;
 	HitoriVector iter;
 	guint board_width, board_height;
 	gfloat cell_size;
 	gdouble x_pos, y_pos;
-	GtkStyle *style;
-	gboolean has_won = FALSE;
+	GtkStyleContext *style_context;
+	GtkBorder border;
 
-	/* Check to see if all three rules are satisfied yet.
-	 * If they are, we've won.
-	 * NOTE: We check rule 1 last, as it's the only rule
-	 * which won't set an error position. */
-	if (check_win &&
-	    hitori_check_rule2 (hitori) &&
-	    hitori_check_rule3 (hitori) &&
-	    hitori_check_rule1 (hitori)) {
-		/* Win! (Tell them later, once we've re-rendered the winning board) */
-		has_won = TRUE;
-	}
-
-	gdk_drawable_get_size (GDK_DRAWABLE (gtk_widget_get_window (hitori->drawing_area)), &area_width, &area_height);
-	style = gtk_widget_get_style (hitori->drawing_area);
+	area_width = gdk_window_get_width (gtk_widget_get_window (hitori->drawing_area));
+	area_height = gdk_window_get_height (gtk_widget_get_window (hitori->drawing_area));
+	style_context = gtk_widget_get_style_context (hitori->drawing_area);
 
 	/* Clamp the width/height to the minimum */
 	if (area_height < area_width) {
@@ -125,7 +131,8 @@ hitori_draw_board (Hitori *hitori, cairo_t *cr, gboolean check_win)
 
 	/* Work out the cell size and scale all text accordingly */
 	cell_size = board_width / hitori->board_size;
-	pango_font_description_set_absolute_size (style->font_desc, cell_size * FONT_SCALE * 0.8 * PANGO_SCALE);
+	pango_font_description_set_absolute_size (hitori->normal_font_desc, cell_size * NORMAL_FONT_SCALE * 0.8 * PANGO_SCALE);
+	pango_font_description_set_absolute_size (hitori->painted_font_desc, cell_size * PAINTED_FONT_SCALE * 0.8 * PANGO_SCALE);
 
 	/* Centre the board */
 	hitori->drawing_area_x_offset = (area_width - board_width) / 2;
@@ -133,33 +140,39 @@ hitori_draw_board (Hitori *hitori, cairo_t *cr, gboolean check_win)
 	cairo_translate (cr, hitori->drawing_area_x_offset, hitori->drawing_area_y_offset);
 
 	/* Draw the cells */
-	x_pos = 0;
-	for (iter.x = 0; iter.x < hitori->board_size; iter.x++) { /* columns (X) */
-		y_pos = 0;
-		for (iter.y = 0; iter.y < hitori->board_size; iter.y++) { /* rows (Y) */
+	for (iter.x = 0, x_pos = 0; iter.x < hitori->board_size; iter.x++, x_pos += cell_size) { /* columns (X) */
+		for (iter.y = 0, y_pos = 0; iter.y < hitori->board_size; iter.y++, y_pos += cell_size) { /* rows (Y) */
 			gchar *text;
 			PangoLayout *layout;
 			gint text_width, text_height;
-			GtkStateType state = GTK_STATE_NORMAL;
+			GtkStateFlags state = 0;
+			gboolean painted = FALSE;
+			PangoFontDescription *font_desc;
+			GdkRGBA colour;
 
-			if (hitori->board[iter.x][iter.y].status & CELL_PAINTED)
-				state = GTK_STATE_INSENSITIVE;
+			if (hitori->board[iter.x][iter.y].status & CELL_PAINTED) {
+				painted = TRUE;
+				state = GTK_STATE_FLAG_INSENSITIVE;
+			}
+
+			/* Set up the border */
+			gtk_style_context_get_border (style_context, state, &border);
+			border.left = 2.0; /* Hack! */
 
 			/* Draw the fill */
-			if (hitori->display_error == TRUE &&
-			    hitori->error_position.x == iter.x && hitori->error_position.y == iter.y)
-				cairo_set_source_rgb (cr, 0.678431373, 0.498039216, 0.658823529); /* Tango's lightest "plum" */
-			else
-				gdk_cairo_set_source_color (cr, &style->bg[state]);
+			if (hitori->debug == TRUE) {
+				g_debug ("State: %u", state);
+			}
+
+			gtk_style_context_get_background_color (style_context, state, &colour);
+			gdk_cairo_set_source_rgba (cr, &colour);
 			cairo_rectangle (cr, x_pos, y_pos, cell_size, cell_size);
 			cairo_fill (cr);
 
 			/* If the cell is tagged, draw the tag dots */
 			if (hitori->board[iter.x][iter.y].status & CELL_TAG1) {
-				if (hitori->board[iter.x][iter.y].status & CELL_PAINTED)
-					cairo_set_source_rgb (cr, 0.643137255, 0, 0); /* Tango's darkest "scarlet red" */				
-				else
-					cairo_set_source_rgb (cr, 0.937254902, 0.160784314, 0.160784314); /* Tango's lightest "scarlet red" */
+				/* Tango's lightest "sky blue" */
+				cairo_set_source_rgba (cr, 0.447058824, 0.623529412, 0.811764706, (painted == TRUE) ? PAINTED_ALPHA : NORMAL_ALPHA);
 
 				cairo_move_to (cr, x_pos, y_pos + TAG_OFFSET);
 				cairo_line_to (cr, x_pos, y_pos);
@@ -169,10 +182,8 @@ hitori_draw_board (Hitori *hitori, cairo_t *cr, gboolean check_win)
 			}
 
 			if (hitori->board[iter.x][iter.y].status & CELL_TAG2) {
-				if (hitori->board[iter.x][iter.y].status & CELL_PAINTED)
-					cairo_set_source_rgb (cr, 0.305882353, 0.603921569, 0.023529412); /* Tango's darkest "chameleon" */
-				else
-					cairo_set_source_rgb (cr, 0.541176471, 0.88627451, 0.203921569); /* Tango's lightest "chameleon" */
+				/* Tango's lightest "chameleon" */
+				cairo_set_source_rgba (cr, 0.541176471, 0.88627451, 0.203921569, (painted == TRUE) ? PAINTED_ALPHA : NORMAL_ALPHA);
 
 				cairo_move_to (cr, x_pos + cell_size - TAG_OFFSET, y_pos);
 				cairo_line_to (cr, x_pos + cell_size, y_pos);
@@ -182,8 +193,9 @@ hitori_draw_board (Hitori *hitori, cairo_t *cr, gboolean check_win)
 			}
 
 			/* Draw the border */
-			gdk_cairo_set_source_color (cr, &style->dark[state]);
-			cairo_set_line_width (cr, style->xthickness);
+			gtk_style_context_get_border_color (style_context, state, &colour);
+			gdk_cairo_set_source_rgba (cr, &colour);
+			cairo_set_line_width (cr, border.left);
 			cairo_rectangle (cr, x_pos, y_pos, cell_size, cell_size);
 			cairo_stroke (cr);
 
@@ -192,81 +204,39 @@ hitori_draw_board (Hitori *hitori, cairo_t *cr, gboolean check_win)
 			layout = pango_cairo_create_layout (cr);
 
 			pango_layout_set_text (layout, text, -1);
-			pango_layout_set_font_description (layout, style->font_desc);
+
+			font_desc = (painted == TRUE) ? hitori->painted_font_desc : hitori->normal_font_desc;
+
+			if (hitori->board[iter.x][iter.y].status & CELL_ERROR) {
+				cairo_set_source_rgb (cr, 0.937254902, 0.160784314, 0.160784314); /* Tango's lightest "scarlet red" */
+				pango_font_description_set_weight (font_desc, PANGO_WEIGHT_BOLD);
+			} else {
+				gtk_style_context_get_color (style_context, state, &colour);
+				gdk_cairo_set_source_rgba (cr, &colour);
+				pango_font_description_set_weight (font_desc, PANGO_WEIGHT_NORMAL);
+			}
+
+			pango_layout_set_font_description (layout, font_desc);
 
 			pango_layout_get_pixel_size (layout, &text_width, &text_height);
 			cairo_move_to (cr,
 				       x_pos + (cell_size - text_width) / 2,
 				       y_pos + (cell_size - text_height) / 2);
 
-			gdk_cairo_set_source_color (cr, &style->text[state]);
 			pango_cairo_show_layout (cr, layout);
 
 			g_free (text);
 			g_object_unref (layout);
-
-			y_pos += cell_size;
 		}
-
-		x_pos += cell_size;
-	}
-
-	if (has_won == TRUE) {
-		GtkWidget *dialog;
-
-		/* Tell the user they've won, and don't draw any hints */
-		hitori_disable_events (hitori);
-		dialog = gtk_message_dialog_new (NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_OK,
-						 _("You've won!"));
-		gtk_dialog_run (GTK_DIALOG (dialog));
-		gtk_widget_destroy (dialog);
-		return;
 	}
 
 	/* Draw a hint if applicable */
 	if (hitori->hint_status % 2 == 1) {
 		cairo_set_source_rgb (cr, 1, 0, 0); /* red */
+		cairo_set_line_width (cr, border.left * 5.0);
 		cairo_rectangle (cr, hitori->hint_position.x * cell_size, hitori->hint_position.y * cell_size, cell_size, cell_size);
 		cairo_stroke (cr);
 	}
-}
-
-/**
- * hitori_draw_board_simple:
- * @hitori: a #Hitori
- * @check_win: %TRUE if the win status should be checked after drawing
- * @clear_first: %TRUE if the drawing area should be cleared before re-drawing
- *
- * Allows easy re-drawing of the board. If @check_win is %TRUE, the board will
- * be checked to see if the player's won after it's been drawn, and if @clear_first
- * is %TRUE, the drawing area will be completely cleared before being drawn.
- * @clear_first is intended to be used only if there's a possibility the size of the
- * board has changed.
- **/
-void
-hitori_draw_board_simple (Hitori *hitori, gboolean check_win, gboolean clear_first)
-{
-	cairo_t *cr;
-
-	if (clear_first)
-		gdk_window_clear (gtk_widget_get_window (hitori->drawing_area));
-
-	cr = gdk_cairo_create (GDK_DRAWABLE (gtk_widget_get_window (hitori->drawing_area)));
-	hitori_draw_board (hitori, cr, check_win);
-	cairo_destroy (cr);
-}
-
-gboolean
-hitori_expose_cb (GtkWidget *drawing_area, GdkEventExpose *event, Hitori *hitori)
-{
-	cairo_t *cr;
-
-	cr = gdk_cairo_create (GDK_DRAWABLE (gtk_widget_get_window (hitori->drawing_area)));
-	cairo_rectangle (cr, event->area.x, event->area.y,
-				event->area.width, event->area.height);
-	cairo_clip (cr);
-	hitori_draw_board (hitori, cr, FALSE);
-	cairo_destroy (cr);
 
 	return FALSE;
 }
@@ -283,7 +253,8 @@ hitori_button_release_cb (GtkWidget *drawing_area, GdkEventButton *event, Hitori
 	if (hitori->processing_events == FALSE)
 		return FALSE;
 
-	gdk_drawable_get_size (GDK_DRAWABLE (gtk_widget_get_window (hitori->drawing_area)), &width, &height);
+	width = gdk_window_get_width (gtk_widget_get_window (hitori->drawing_area));
+	height = gdk_window_get_height (gtk_widget_get_window (hitori->drawing_area));
 
 	/* Clamp the width/height to the minimum */
 	if (height < width)
@@ -306,7 +277,12 @@ hitori_button_release_cb (GtkWidget *drawing_area, GdkEventButton *event, Hitori
 	undo->undo = hitori->undo_stack;
 	undo->redo = NULL;
 
-	if (event->state & GDK_SHIFT_MASK) {
+	if (event->state & GDK_SHIFT_MASK && event->state & GDK_CONTROL_MASK) {
+		/* Update both tags' state */
+		hitori->board[pos.x][pos.y].status ^= CELL_TAG1;
+		hitori->board[pos.x][pos.y].status ^= CELL_TAG2;
+		undo->type = UNDO_TAGS;
+	} else if (event->state & GDK_SHIFT_MASK) {
 		/* Update tag 1's state */
 		hitori->board[pos.x][pos.y].status ^= CELL_TAG1;
 		undo->type = UNDO_TAG1;
@@ -329,13 +305,14 @@ hitori_button_release_cb (GtkWidget *drawing_area, GdkEventButton *event, Hitori
 	gtk_action_set_sensitive (hitori->undo_action, TRUE);
 
 	/* Stop any current hints */
-	hitori->hint_status = HINT_FLASHES;
-
-	if (hitori->debug)
-		g_debug ("Stopping all current hints.");
+	hitori_cancel_hinting (hitori);
 
 	/* Redraw */
-	hitori_draw_board_simple (hitori, recheck, FALSE);
+	gtk_widget_queue_draw (hitori->drawing_area);
+
+	/* Check to see if the player's won */
+	if (recheck == TRUE)
+		hitori_check_win (hitori);
 
 	return FALSE;
 }
@@ -346,31 +323,60 @@ hitori_destroy_cb (GtkWindow *window, Hitori *hitori)
 	hitori_quit (hitori);
 }
 
+void
+hitori_window_state_event_cb (GtkWindow *window, GdkEventWindowState *event, Hitori *hitori)
+{
+	if (hitori->debug)
+		g_debug ("Got window state event: %u (changed: %u)", event->new_window_state, event->changed_mask);
+
+	if (event->new_window_state & GDK_WINDOW_STATE_WITHDRAWN ||
+	    event->new_window_state & GDK_WINDOW_STATE_ICONIFIED ||
+	    event->new_window_state & GDK_WINDOW_STATE_BELOW) {
+		/* Pause the timer */
+		hitori_pause_timer (hitori);
+	} else {
+		/* Re-start the timer */
+		hitori_start_timer (hitori);
+	}
+}
+
 void
 hitori_new_game_cb (GtkAction *action, Hitori *hitori)
 {
 	hitori_new_game (hitori, hitori->board_size);
 }
 
+static void
+hitori_cancel_hinting (Hitori *hitori)
+{
+	if (hitori->debug)
+		g_debug ("Stopping all current hints.");
+
+	hitori->hint_status = HINT_DISABLED;
+	if (hitori->hint_timeout_id != 0)
+		g_source_remove (hitori->hint_timeout_id);
+	hitori->hint_timeout_id = 0;
+}
+
 static gboolean
 hitori_update_hint (Hitori *hitori)
 {
-	cairo_t *cr;
 	gint area_width, area_height;
-	guint board_width, board_height;
+	guint board_width, board_height, drawing_area_x_offset, drawing_area_y_offset;
 	gfloat cell_size;
 
 	/* Check to see if hinting's been stopped by a cell being changed (race condition) */
-	if (hitori->hint_status >= HINT_FLASHES)
+	if (hitori->hint_status == HINT_DISABLED)
 		return FALSE;
 
-	hitori->hint_status++;
+	hitori->hint_status--;
 
 	if (hitori->debug)
 		g_debug ("Updating hint status to %u.", hitori->hint_status);
 
 	/* Calculate the area to redraw (just the hinted cell, hopefully) */
-	gdk_drawable_get_size (GDK_DRAWABLE (gtk_widget_get_window (hitori->drawing_area)), &area_width, &area_height);
+	area_width = gdk_window_get_width (gtk_widget_get_window (hitori->drawing_area));
+	area_height = gdk_window_get_height (gtk_widget_get_window (hitori->drawing_area));
 
 	/* Clamp the width/height to the minimum */
 	if (area_height < area_width) {
@@ -383,23 +389,19 @@ hitori_update_hint (Hitori *hitori)
 
 	cell_size = board_width / hitori->board_size;
 
-	/* Centre the board */
-	cr = gdk_cairo_create (GDK_DRAWABLE (gtk_widget_get_window (hitori->drawing_area)));
-	cairo_save (cr);
+	drawing_area_x_offset = (area_width - board_width) / 2;
+	drawing_area_y_offset = (area_height - board_height) / 2;
 
-	hitori->drawing_area_x_offset = (area_width - board_width) / 2;
-	hitori->drawing_area_y_offset = (area_height - board_height) / 2;
-	cairo_translate (cr, hitori->drawing_area_x_offset, hitori->drawing_area_y_offset);
+	/* Redraw the cell */
+	gtk_widget_queue_draw_area (hitori->drawing_area, drawing_area_x_offset + hitori->hint_position.x * cell_size,
+	                            drawing_area_y_offset + hitori->hint_position.y * cell_size, cell_size, cell_size);
 
-	/* Clip to the cell */
-	cairo_rectangle (cr, hitori->hint_position.x * cell_size, hitori->hint_position.y * cell_size, cell_size, cell_size);
-	cairo_clip (cr);
-	cairo_restore (cr);
+	if (hitori->hint_status == HINT_DISABLED) {
+		hitori_cancel_hinting (hitori);
+		return FALSE;
+	}
 
-	hitori_draw_board (hitori, cr, FALSE);
-	cairo_destroy (cr);
-
-	return (hitori->hint_status < HINT_FLASHES) ? TRUE : FALSE;
+	return TRUE;
 }
 
 void
@@ -407,6 +409,10 @@ hitori_hint_cb (GtkAction *action, Hitori *hitori)
 {
 	HitoriVector iter;
 
+	/* Bail if we're already hinting */
+	if (hitori->hint_status != HINT_DISABLED)
+		return;
+
 	/* Find the first cell which should be painted, but isn't (or vice-versa) */
 	for (iter.x = 0; iter.x < hitori->board_size; iter.x++) {
 		for (iter.y = 0; iter.y < hitori->board_size; iter.y++) {
@@ -418,9 +424,9 @@ hitori_hint_cb (GtkAction *action, Hitori *hitori)
 					g_debug ("Beginning hinting in cell (%u,%u).", iter.x, iter.y);
 
 				/* Set up the cell for hinting */
-				hitori->hint_status = 0;
+				hitori->hint_status = HINT_FLASHES;
 				hitori->hint_position = iter;
-				g_timeout_add (HINT_INTERVAL, (GSourceFunc) hitori_update_hint, hitori);
+				hitori->hint_timeout_id = g_timeout_add (HINT_INTERVAL, (GSourceFunc) hitori_update_hint, hitori);
 				hitori_update_hint ((gpointer) hitori);
 
 				return;
@@ -445,6 +451,10 @@ hitori_undo_cb (GtkAction *action, Hitori *hitori)
 		case UNDO_TAG2:
 			hitori->board[hitori->undo_stack->cell.x][hitori->undo_stack->cell.y].status ^= CELL_TAG2;
 			break;
+		case UNDO_TAGS:
+			hitori->board[hitori->undo_stack->cell.x][hitori->undo_stack->cell.y].status ^= CELL_TAG1;
+			hitori->board[hitori->undo_stack->cell.x][hitori->undo_stack->cell.y].status ^= CELL_TAG2;
+			break;
 		case UNDO_NEW_GAME:
 		default:
 			/* This is just here to stop the compiler warning */
@@ -458,8 +468,11 @@ hitori_undo_cb (GtkAction *action, Hitori *hitori)
 	if (hitori->undo_stack->undo == NULL || hitori->undo_stack->type == UNDO_NEW_GAME)
 		gtk_action_set_sensitive (hitori->undo_action, FALSE);
 
+	/* The player can't possibly have won, but we need to update the error highlighting */
+	hitori_check_win (hitori);
+
 	/* Redraw */
-	hitori_draw_board_simple (hitori, TRUE, FALSE);
+	gtk_widget_queue_draw (hitori->drawing_area);
 }
 
 void
@@ -480,6 +493,10 @@ hitori_redo_cb (GtkAction *action, Hitori *hitori)
 		case UNDO_TAG2:
 			hitori->board[hitori->undo_stack->cell.x][hitori->undo_stack->cell.y].status ^= CELL_TAG2;
 			break;
+		case UNDO_TAGS:
+			hitori->board[hitori->undo_stack->cell.x][hitori->undo_stack->cell.y].status ^= CELL_TAG1;
+			hitori->board[hitori->undo_stack->cell.x][hitori->undo_stack->cell.y].status ^= CELL_TAG2;
+			break;
 		case UNDO_NEW_GAME:
 		default:
 			/* This is just here to stop the compiler warning */
@@ -491,8 +508,11 @@ hitori_redo_cb (GtkAction *action, Hitori *hitori)
 	if (hitori->undo_stack->redo == NULL)
 		gtk_action_set_sensitive (hitori->redo_action, FALSE);
 
+	/* The player can't possibly have won, but we need to update the error highlighting */
+	hitori_check_win (hitori);
+
 	/* Redraw */
-	hitori_draw_board_simple (hitori, TRUE, FALSE);
+	gtk_widget_queue_draw (hitori->drawing_area);
 }
 
 void
@@ -507,7 +527,7 @@ hitori_contents_cb (GtkAction *action, Hitori *hitori)
 	GError *error = NULL;
 
 	if (gtk_show_uri (gtk_widget_get_screen (hitori->window), "ghelp:hitori", gtk_get_current_event_time (), &error) == FALSE) {
-		GtkWidget *dialog = gtk_message_dialog_new (NULL,
+		GtkWidget *dialog = gtk_message_dialog_new (GTK_WINDOW (hitori->window),
 							    GTK_DIALOG_MODAL,
 							    GTK_MESSAGE_ERROR,
 							    GTK_BUTTONS_OK,
